/*
 * Author: Giam Teck Choon <choon@choon.net>
 *
 * Description: mailfront-imapfront-auth-dovecot is a program to modify the
 * MAIL environment variable according to dovecot imap requirements in order
 * to access users mail storage.
 * 
 * Dovecot uses colons as internal field separators; however, since v1.2rc6, it
 * allows colons occurrences via escaping (as : → ::). Before passing such
 * mailbox paths via MAIL environment variable to dovecot imap, this wrapper
 * performs such escaping.
 * 
 * From original MAIL environment variable:
 * 
 * /home/USERNAME/MAILDIR/mail:box:user
 * 
 * To:
 * 
 * maildir:/home/USERNAME/MAILDIR/mail::box::user
 * 
 * More information can be read at:
 * 
 * http://wiki.dovecot.org/MailLocation
 * 
 * References:
 * 
 * http://mij.oltrelinux.com/net/dovecot-qmail-vmailmgr/
 * http://wiki.dovecot.org/HowTo/VMailMgr
 *
 * Copyright: This package is Copyright(C) 2010 Giam Teck Choon or CHOON.NET, and may be copied
 *            according to the GNU GENERAL PUBLIC LICENSE (GPL) Version 2 or a later version. A
 *            copy of this license is included with this package. This package comes with no
 *            warranty of any kind.
 *
 * Disclaimer: Use this program at your own risk! I am not responsible for any damages caused.
 *             You have been warned!
 *
 * Code reference:
 * http://mij.oltrelinux.com/net/dovecot-qmail-vmailmgr/checkvpw-dovecot-wrapper.c
 * MANY THANKS TO mij@bitchx.it
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#define PROGNAME                "mailfront-imapfront-auth-dovecot"
/* what maximum length we can expect in MAIL environment variable */
#define MAXLEN                  600

/* replace all ":" with "::" in string. Return if a replacement occurred. */
int escape_colon(char *string, size_t buflen) {
    int counter, numcolons;

    /* any ':' in original string? */
    for (numcolons = counter = 0; string[counter] != '\0'; counter++) {
        if (string[counter] == ':') numcolons++;
    }
    if (numcolons > 0) {
        char *tmpstr;
        size_t newlen;
        int j; /* index for iteration over second string */

        /* copy string -> tmpstr, duplicating all ':' -> '::' */
        newlen = strlen(string) + numcolons + 1;
        if (newlen > buflen) {
	    return 111;
        }
        tmpstr = (char *)malloc(newlen);
        /* copy everything, but put "::" when ":" is found */
        for (j = counter = 0; string[counter] != '\0'; counter++, j++) {
            tmpstr[j] = string[counter];
            if (string[counter] == ':') {
                /* add another ':' */
                j++;
                tmpstr[j] = ':';
            }
        }
        tmpstr[j] = '\0';

        /* now copy over to original string */
        strcpy(string, tmpstr);
        free(tmpstr);

        return 1;
    }

    return 0;
}

int main(int argc, char *argv[]) {
    char buffer[MAXLEN];
    unsigned long pid;
    const char *ip;
    const char *tmp;
    char *maildir = "maildir:";
    char mailenv[MAXLEN+10];
    int debug = 0;

    if (argc < 2) {
	fprintf(stderr, "usage: %s program [arguments]\n", PROGNAME);
	exit(111);
    }

    pid = (unsigned long)getpid();

    /* get IP from environment variable TCPREMOTEIP */
    if (getenv("IMAPFRONTAUTH_DOVECOT_DEBUG") != 0) {
	debug = 1;
    }

    /* get IP from environment variable TCPREMOTEIP */
    if (getenv("TCPREMOTEIP") != 0) {
	ip = getenv("TCPREMOTEIP");
    } else {
	ip = "";
    }

    /* check for environment variable MAIL */
    tmp = getenv("MAIL");
    if (strlen(tmp) > 0) {
	if (strlen(tmp) > MAXLEN) {
	    if (debug > 0) {
		fprintf(stderr, "%s: %s pid %lu: MAIL environment variable length is greater than %d!\n",PROGNAME, ip, pid, MAXLEN);
		printf("MAIL environment variable is too long!\r\n");
	    }
	    exit(111);
	} else {
	    /* escape possible ':' characters before finally exporting */
	    strcpy(buffer, tmp);
	    if (escape_colon(buffer, MAXLEN) == 111) {
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: buffer too short to accommodate the resulting MAIL environment\n", PROGNAME, ip, pid);
		    printf("buffer too short to accommodate the resulting MAIL environment!\r\n");
		}
		exit(111);
	    } else {
		strcat(mailenv, maildir);
		strcat(mailenv, buffer);
		setenv("MAIL", mailenv, 1);
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: MAIL environment is %s\n", PROGNAME, ip, pid, mailenv);
		    printf("MAIL environment is %s\r\n", mailenv);
		}
		execvp(argv[1], argv+1);
		fprintf(stderr, "%s: %s pid %lu: unable to run %s: %s\n",PROGNAME, ip, pid, argv[1], strerror(errno));
		exit(111);
	    }
	}
    } else {
	fprintf(stderr, "%s: %s pid %lu: MAIL environment variable not found!\n",PROGNAME, ip, pid);
	printf("MAIL environment variable is required!\r\n");
	exit(111);
    }
}
